from typing import Iterable, Self, Tuple, Hashable
from collections.abc import MutableSet, Set
from abc import ABC, abstractmethod, abstractproperty

class Symbol:
    """
    Ex:
    ===
        >>> A = Symbol("Some Name", 2)
        >>> print(A)
        Symb(Some Name, 2)
        >>> A(3, 4) # this is a fact

    """

    def __init__(self, name: str, arity: int):
        self.arity = arity
        self.name = name

    def __call__(self, *args) -> "Atom | Fact":
        assert len(args) == self.arity
        if any(isinstance(a, v) for a in args):
            return Atom(self, args)
        return Fact(self, args)

    def __eq__(self, other):
        return type(self) == type(other) and self.name == other.name and self.arity == other.arity

    def __hash__(self):
        return hash((self.arity, self.name))

    def __repr__(self):
        return f"Symb({self.name}, {self.arity})"


class v(str):
    """
    Ex:
    ===
        >>> x = v("x")
        >>> x2 = v("x")
        >>> x == x2
        True
    """

    def __eq__(self, other):
        return type(self) is type(other) and super().__eq__(other)

    def __repr__(self):
        return f"v({self})"

    def __hash__(self):
        return super().__hash__()


class Atom:
    """
    Ex:
    ===
        >>> A = Symbol("Some Name", 3)
        >>> Atom(A, (3, 4, v("x")))
        A(3, 4, v("x")) # This is the Atom(3, 4, x) with x a variable
        
    """
    def __init__(self, symbol: Symbol, data: tuple): 
        self.symbol = symbol
        self.data = data

    def __hash__(self):
        return hash((self.symbol, self.data))

    def __repr__(self):
        return f"{self.symbol}{self.data}"


class Fact(Atom):
    def __init__(self, symbol: Symbol, data: tuple): 
        super().__init__(symbol, data)
        for val in data:
            # DO NOT DO type(val) == v
            if isinstance(val, v):
                raise ValueError(f"A fact do not contains variable, got {val}")


        # if any(isinstance(val, v) for var in data):
        #     raise ValueError(...)
        # TODO


class CQ:
    """
    Ex:
    ===
        # Q(x, x, y) := Exists z R(x, y, z)
        >>> R = Symbol("R", 3)
        >>> CQ((v("x"), v("x"), v("y")), R(v("x"), v("y"), v("z")))

        # Q(x) := Exists z,  Q(x, "a") and V("a", "b", z)

        >>> V = Symbol("V", 3)
        >>> Q = Symbol("Q", 2)
        >>> CQ((v("x"),), Q(v("x"),"a"), V("a", "b", v("z")))

    """
    def __init__(self, out_variables: tuple[v], *atoms: tuple[Atom, ...]):
        self.out_variables = out_variables
        self.atoms = atoms

    @property
    def variables(self) -> frozenset[v]:
        S = set()
        # S = frozenset() BAD
        for atom in self.atoms:
            for val in atom.data:
                if isinstance(val, v):
                    S.add(val)
                    # S = S.union([val]) BAD

        # BAD is O(n^2) while code is O(n)
        return frozenset(S)

        def generator():
            yield from (val for atom in self.atoms 
                            for val in atom.data if isinstance(val, v))
        return frozenset(generator()) # This is even better !

    @property
    def is_boolean(self):
        return len(self.out_variables) == 0
        
    def containment(self, query: "CQ") -> bool:
        ... # TODO

    def substitute(self, var: v, value) -> "CQ":
        new_out_variable = tuple((nvar for nvar in self.out_variables 
                                  if nvar != var))
        new_atoms = []
        for atom in self.atoms:
            data = tuple((value if val == var else val for val in atom.data))
            new_atoms.append(Atom(atom.symbol, data))
        return CQ(new_out_variable, *new_atoms)




class UCQ(frozenset[CQ]): ...

class Database(ABC, Set[Fact]):
    @abstractproperty
    def relations(self) -> frozenset[Symbol]: 
        """ Give all the symbol occuring in the Database """
        ...

    @abstractmethod
    def naive_evaluation(self, query: CQ | UCQ) -> frozenset[tuple]:
        ...

    @abstractmethod
    def sat_evaluation(self, query) -> frozenset[tuple]:
        ...

    @abstractproperty
    def active_domain(self) -> frozenset:
        ...

    def __repr__(self):
        return f"DB({','.join(map(str, self))})"

    @abstractmethod
    def relation(self, rel: Symbol) -> frozenset[tuple[Fact]]:
        """ Returns all the tuple associated to `rel` symbol"""

class StaticDatabase(frozenset[Fact], Database):
    """ This class should implement database while being a frozenset.

        Ex:
        ===
            >>> A = Symbol("A", 2)
            >>> B = Symbol("B", 1)
            >>> DB = StaticDatabase([A(1, 2), B(2), A(3, 4)])
            >>> DB.relations

            >>> DB.relation(A)
            frozenset({A(1, 2), A(3, 4)})
            >>> DB.relation(B)
            frozenset({B(2)})
    """
    @property
    def relations(self) -> frozenset[Symbol]: 
        """ Give all the symbol occuring in the Database """
        # S = set()
        # for fact in self:
        #    S.add(fact.symbol)

        # alternative 

        return frozenset((fact.symbol for fact in self))
        # return frozenset([fact.symbol for fact in self]) BAD


    def naive_evaluation(self, query: CQ | UCQ) -> frozenset[tuple]:
        ...

    @abstractmethod
    def sat_evaluation(self, query) -> frozenset[tuple]:
        ...

    def active_domain(self) -> frozenset:
        return frozenset((d for fact in self for d in fact.data))
        

    def __repr__(self):
        return f"DB({','.join(map(str, self))})"

    def relation(self, rel: Symbol) -> frozenset[tuple]:
        """ Returns all the tuple associated to `rel` symbol"""
        return frozenset((fact.data for fact in self if fact.symbol == rel))

class DynamicDatabase(Database, MutableSet[Fact]):
    """ This class is dynamic and fact can be update or delete 
        Try to implement a Database with the lower complexity
        when updating is performed !

        To implement MutableSet you need to implement the following function:

        * __contains__: check if a fact is in the database

        * __iter__: iter on all facts 

        * __len__: give the number of facts

        * add: add a new fact

        * discard: delete a fact if it is the database or do nothing otherwise

    """
    def __init__(self):
        self.datastore : set[fact] = set()  # We store the fact here !



if __name__ == "__main__":
    Lectures = Symbol("Lecture", 2)
    Person = Symbol("Person", 3)
    charles = Person(0, "Charles", "Paperman")
    mikael = Person(1, "Mikaël", "Monet")
    DB1 = Lectures("Database", 1)
    AC2 = Lectures("AC2", 0)
    SD = StaticDatabase([charles, mikael, AC2, DB1])

